/*
** File: platform.c  
** Project: ADMXRC2 API module library
** Purpose: Implements VxWorks-specific functions for ADMXRC2 API module library
**
** (C) Copyright Alpha Data 2016
**
** Mechanism for reporting errors from driver to API when calling ioctl() is as follows:
**
**   - A successful ioctl() call always returns OK.
**   - An error reported by the driver framework results in ioctl() returning ERROR, and the driver setting errno appropriately. See below for
**     errno usage.
**   - An error reported by the platform-independent code results in ioctl() returning a positive number, which is then cast to ADMXRC2_STATUS.
**
** Error numbers (errno) are used as follows:
**
**   EACCES      (O)    - Insufficient rights to open device; shouldn't happen in VxWorks
**   ENXIO       (O)    - Device not found (i.e. /admxrc2/<index> doesn't exist
**   ENDEV       (O)    - As ENXIO
**   ENOMEM      (O, I) - Couldn't allocate memory, either from heap or from a pool in order to perform operation
**   EBADF       (I, C) - File descriptor was not valid
**   EFAULT      (I)    - The IOCTL structure passed was NULL, or the in/out size encoded in the IOCTL number was not what the driver expected.
**   ENOSYS      (I)    - The IOCTL number passed was not one recognized by the driver
**   EBUSY       (I)    - There is already a non-blocking operation in progress
**   EINPROGRESS (I)    - The non-blocking operation is still ongoing
**   EPIPE       (I)    - No non-blocking operation to finish
**
** where O, I, and C mean applicable to open(), ioctl(), and close() respectively.
*/

#include <ioLib.h>
#include <selectLib.h>
#include <assert.h>
#include <errno.h>
#include <string.h>
#if ADMXRC2_VXWORKS_RTP
# include <sys/mman.h>
#endif

#if (_WRS_VXWORKS_MAJOR > 6) || (_WRS_VXWORKS_MAJOR == 6 && _WRS_VXWORKS_MINOR >= 9)
# define HAS_VX_IOCTL_ARG_T (1)
#else
# define HAS_VX_IOCTL_ARG_T (0)
#endif

#include "admxrc2_platform.h"
#include <ioctl_dfvxw.h>

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

static const char g_deviceName[] = "admxrc2i";

ADMXRC2_STATUS
admxrc2OpenDevice(
  unsigned int index,
  bool_t bPassive, /* ignored in VxWorks */
  ADMXRC2_HANDLE* phAdmxrc2)
{
  char filename[32];
  int hDevice;

#if (_WRS_VXWORKS_MAJOR >= 6)
  snprintf(filename, ARRAY_LENGTH(filename), "/%s/%u", g_deviceName, index);
#else
  /* Need VxWorks 6.x for snprintf() */
  sprintf(filename, "/%s/%u", g_deviceName, index);
#endif
  hDevice = open(filename, O_RDWR, 0);
  if (ERROR == hDevice) {
    switch (errno) {
    case EACCES:
      return ADMXRC2_ACCESS_DENIED;

    case ENOMEM:
      return ADMXRC2_NO_MEMORY;

    case ENXIO:
    case ENODEV:
      return ADMXRC2_CARD_NOT_FOUND;

    default:
      fprintf(stderr, "admxrc2OpenDevice: errno=%d (%s)\n", errno, strerror(errno));
      return ADMXRC2_UNKNOWN_ERROR;
    }
  }
  *phAdmxrc2 = hDevice;
  return ADMXRC2_SUCCESS;
}

ADMXRC2_STATUS
admxrc2CloseDevice(
  ADMXRC2_HANDLE hAdmxrc2)
{
  admxrc2UnmapAllRegions(hAdmxrc2);

  if (ERROR == close(hAdmxrc2)) {
    switch (errno) {
      case EBADF:
        return ADMXRC2_INVALID_HANDLE;

      default:
        fprintf(stderr, "admxrc2CloseDevice: errno=%d (%s)\n", errno, strerror(errno));
        return ADMXRC2_UNKNOWN_ERROR;
    }
  }    

  return ADMXRC2_SUCCESS;
}

ADMXRC2_STATUS
_admxrc2Ioctl(
  ADMXRC2_HANDLE hAdmxrc2,
  unsigned int code,
  void* pIoctl,
  unsigned int sizeIn,
  unsigned int sizeOut)
{
  STATUS err;

  if (sizeIn >= DFVXW_MAX_IOCTL_SIZE || sizeOut >= DFVXW_MAX_IOCTL_SIZE) {
    fprintf(stderr, "admxrc2Ioctl: sizeIn (%u) or sizeOut (%u) greater than 64\n", sizeIn, sizeOut);
    return ADMXRC2_UNKNOWN_ERROR;
  }
#if HAS_VX_IOCTL_ARG_T
  err = ioctl(hAdmxrc2, DFVXW_MAKE_IOCTL(code, sizeIn, sizeOut), (_Vx_ioctl_arg_t)pIoctl);
#else
  err = ioctl(hAdmxrc2, DFVXW_MAKE_IOCTL(code, sizeIn, sizeOut), (int)pIoctl);
#endif
  if (OK == err) {
    return ADMXRC2_SUCCESS;
  } else {
    if (ERROR == err) {
      switch (errno) {
      case EBADF:
        return ADMXRC2_INVALID_HANDLE;

      case ENOMEM:
        return ADMXRC2_NO_MEMORY;

      case EINPROGRESS:
        fprintf(stderr, "admxrc2Ioctl: EINPROGRESS returned for blocking operation\n");
        return ADMXRC2_PENDING;

      case ECANCELED:
        return ADMXRC2_CANCELLED;

      case EFAULT:
      case ENOSYS:
      default:
        fprintf(stderr, "admxrc2Ioctl: errno=%d (%s)\n", errno, strerror(errno));
        return ADMXRC2_UNKNOWN_ERROR;
      }
    } else {
      return (ADMXRC2_STATUS)(err & 0xfffU);
    }
  }
}

void
admxrc2ApiInit(
  void)
{
  admxrc2MmapInitialize();
}

/*
** Reference a symbol from each C source module that we care about, in order to prevent
** VxWorks tools from optimizing unreachable code out of the kernel image.
*/
extern unsigned int ADMXRC2_OpenCard(unsigned int, int*); /* admxrc2.c */

void* admxrc2ApiKeepSymbols[] = {
  (void*)&ADMXRC2_OpenCard
};
